from FQE_PG_utils import *
from PG_utils import process


def FQE_PG(env, features, next_states, rewards, policy_net, feature, H, d, m, sample_s=100, lam=0.0):
    # initialize
    features = process(features, d)
    next_states = process(next_states, policy_net.n_state)
    rewards = process(rewards)

    I_m, I_d = np.eye(m), np.eye(d)
    w_1, W_2 = np.zeros(d), np.zeros((d, m))  # w_1 for Q, w_2 for grad Q
    w_r = np.zeros(d)

    Sigma_hat_inv = lam * I_d
    M = np.zeros((d, d))
    grad_M = np.zeros((d, d * m))
    for i in range(np.shape(features)[0]):
        int_vec = np.zeros(d)
        curr_feature = features[i, :]
        next_state = next_states[i, :]
        reward = rewards[i, 0]
        probs = policy_net.get_action(next_state, sample=False)
        for a in range(policy_net.n_action):
            grad = grad_log(policy_net, next_state, a)
            next_feature = feature(next_state, a)
            pi_a = probs[a]
            int_vec += next_feature * pi_a
            for j in range(m):
                grad_M[:, j * d:(j + 1) * d] += np.outer(curr_feature, next_feature * pi_a) * grad[j]
        M += np.outer(curr_feature, int_vec)
        w_r += reward * curr_feature
        Sigma_hat_inv += np.outer(curr_feature, curr_feature)
    Sigma_hat_inv = np.linalg.pinv(Sigma_hat_inv)
    w_r = np.dot(Sigma_hat_inv, w_r)
    M = np.dot(Sigma_hat_inv, M)
    grad_M = np.dot(Sigma_hat_inv, grad_M)

    for h in range(H, 0, -1):
        W_2 = np.dot(grad_M, np.kron(I_m, (w_1[np.newaxis]).T)) + np.dot(M, W_2)
        w_1 = w_r + np.dot(M, w_1)

    # integrate
    result = np.zeros(m)
    result_Q = 0
    for _ in range(sample_s):
        s = env.reset()
        probs = policy_net.get_action(s, sample=False)
        for a in range(policy_net.n_action):
            result += probs[a] * (
                        np.dot(feature(s, a), w_1) * grad_log(policy_net, s, a) + np.dot(W_2.T, feature(s, a)))
            result_Q += probs[a] * np.dot(feature(s, a), w_1)
    return result / sample_s, result_Q / sample_s, Sigma_hat_inv


def FQE(env, D, target_policy, feature, H, d, m, nA, sample_s=100, lam=0.0):
    # initialize
    D = flatten(D)
    w_1 = np.zeros(d)
    w_r = np.zeros(d)

    Sigma_hat_inv = lam * np.eye(d)
    for pair in D:
        curr_feature = feature(pair[0], pair[1])
        w_r += pair[3] * curr_feature
        Sigma_hat_inv += np.outer(curr_feature, curr_feature)
    Sigma_hat_inv = np.linalg.pinv(Sigma_hat_inv)
    w_r = np.dot(Sigma_hat_inv, w_r)

    for h in range(H, 0, -1):
        w_1_2 = np.zeros(d)
        for pair in D:
            next_state = pair[2]
            curr_feature = feature(pair[0], pair[1])
            tmp = 0.0
            for a in range(nA):
                tmp += np.dot(feature(next_state, a), w_1) * target_policy(next_state)[a]
            w_1_2 += tmp * curr_feature
        w_1_2 = np.dot(Sigma_hat_inv, w_1_2)
        w_1 = w_r + w_1_2

    # integrate
    result = 0
    sample_s_i = 0
    while sample_s_i < sample_s:
        s = env.reset()
        for a in range(nA):
            result += target_policy(s)[a] * (np.dot(feature(s, a), w_1))
        sample_s_i += 1
    return result / sample_s
