import numpy as np
import torch
import pickle
from torch.autograd import Variable
from PG_utils import get_flat_grads_from, process, get_features
import scipy
from scipy.linalg import sqrtm
from PG import collect_data

def flatten(lst):
    result = []
    for sublst in lst:
        result.extend(sublst)
    return result

def net_param_num(model):
    result = 0
    for param in model.parameters():
        if param.requires_grad:
            result += param.numel()
    return result

def load_result(filename):
    f = open(filename, 'rb')
    result = pickle.load(f)
    f.close()
    return result

def store_result(obj, filename):
    f = open(filename, 'wb')
    pickle.dump(obj, f)
    f.close()

def make_epsilon_greedy_policy(policy_net, epsilon, nA, nS):
    if epsilon is None:
        def policy_fn(observation):
            with torch.no_grad():
                action_probs = policy_net(Variable(torch.tensor([wrap(observation, nS)]).float()))[0]
            return action_probs.numpy()
    else:
        def policy_fn(observation):
            with torch.no_grad():
                action_probs = policy_net(Variable(torch.tensor([wrap(observation, nS)]).float()))[0]
            A = action_probs.numpy()
            A = (1.0 - epsilon) * A
            B = np.ones(nA, dtype=float) * epsilon / nA
            result = A + B
            result[-1] = 1 - np.sum(result[:-1])
            return result
    return policy_fn

def grad_log(policy_net, s, a):
    policy_net.zero_grad()
    probs = policy_net(Variable(torch.from_numpy(np.reshape(s, [1, policy_net.n_state])).float()))[0]
    selected_probs = probs[a]
    loss = torch.log(selected_probs)
    loss.backward()
    return get_flat_grads_from(policy_net)

def compute_mismatch_dep(env, feature, policy_net, H, d, Sigma_hat_inv_lst, sample_K=50000):
    states, actions, next_states, rewards = collect_data(env, policy_net, sample_K, H, 0)
    features = get_features(states, actions, feature)
    features = process(features, d)
    Sigma = np.zeros((d,d))
    for i in range(np.shape(features)[0]):
        curr_feature = features[i, :]
        Sigma += np.outer(curr_feature, curr_feature)
    Sigma_sq = sqrtm(Sigma)

    return [np.linalg.cond(Sigma_sq @ Sigma_hat_inv @ Sigma_sq) for Sigma_hat_inv in Sigma_hat_inv_lst]

def compute_mismatch(env, feature, policy_net, H, d, lam=0, sample_K=10000):
    states, actions, next_states, rewards = collect_data(env, policy_net, sample_K, H, 0)
    features = get_features(states, actions, feature)
    features = process(features, d)
    Sigma_hat_inv = lam * np.eye(d)
    for i in range(np.shape(features)[0]):
        curr_feature = features[i, :]
        Sigma_hat_inv += np.outer(curr_feature, curr_feature)
    Sigma_hat_inv = np.linalg.pinv(Sigma_hat_inv)
    Sigma_hat_inv_sq = sqrtm(Sigma_hat_inv)

    result1, result2 = 0, []
    for h in range(H):
        nu = np.zeros(d)
        for episode in features:
            nu += episode[h]
        nu /= len(D)
        result1 += ((H - h) ** 2) * np.linalg.norm(np.dot(Sigma_hat_inv_sq, nu)) # ||sigma_inv_sq nu_h||
    return result1, result2
