import numpy as np
import torch
import pickle
from torch.autograd import Variable
from PG_utils import get_flat_grads_from


def flatten(lst):
    result = []
    for sublst in lst:
        result.extend(sublst)
    return result

def net_param_num(model):
    result = 0
    for param in model.parameters():
        if param.requires_grad:
            result += param.numel()
    return result

def load_result(filename):
    f = open(filename, 'rb')
    result = pickle.load(f)
    f.close()
    return result

def store_result(obj, filename):
    f = open(filename, 'wb')
    pickle.dump(obj, f)
    f.close()

def make_epsilon_greedy_policy(policy_net, epsilon, nA, nS):
    if epsilon is None:
        def policy_fn(observation):
            with torch.no_grad():
                action_probs = policy_net(Variable(torch.tensor([wrap(observation, nS)]).float()))[0]
            return action_probs.numpy()
    else:
        def policy_fn(observation):
            with torch.no_grad():
                action_probs = policy_net(Variable(torch.tensor([wrap(observation, nS)]).float()))[0]
            A = action_probs.numpy()
            A = (1.0 - epsilon) * A
            B = np.ones(nA, dtype=float) * epsilon / nA
            result = A + B
            result[-1] = 1 - np.sum(result[:-1])
            return result
    return policy_fn

def grad_log(policy_net, s, a):
    policy_net.zero_grad()
    probs = policy_net(Variable(torch.from_numpy(np.reshape(s, [1, policy_net.n_state])).float()))[0]
    selected_probs = probs[a]
    loss = torch.log(selected_probs)
    loss.backward()
    return get_flat_grads_from(policy_net)
