import numpy as np
import torch
import torch.nn as nn
from torch.autograd import Variable
import sklearn.pipeline
import sklearn.preprocessing
from sklearn.kernel_approximation import RBFSampler


class FeatureNet(nn.Module):
    def __init__(self, s_dim, nA, d):
        super(FeatureNet, self).__init__()
        self.fc1 = nn.Linear(s_dim * nA, 128)
        self.fc2 = nn.Linear(128, d)

    def forward(self, x):
        x = torch.tanh(self.fc1(x))
        x = self.fc2(x)
        return x


def create_tabular_feature(nS, nA):
    d = nS * nA

    def feature(s, a):
        result = np.zeros(d)
        result[int(a) * nS + np.argmax(s)] = 1
        return result

    return feature, d


def create_RBF_feature(env, d, policy, n=10000):
    s_dim = env.observation_space.shape[0]
    nA = env.action_space.n
    curr_d = d / 4
    observation_examples = []
    i = 0
    while i < n:
        obs_example = np.zeros(s_dim * nA)
        s_tmp = env.reset()
        obs_probs = policy(s_tmp)
        a = np.random.choice(np.arange(len(obs_probs)), p=obs_probs)
        obs_example[a * s_dim: (a + 1) * s_dim] = s_tmp
        observation_examples.append(obs_example)
        i += 1
    observation_examples = np.array(observation_examples)
    scaler = sklearn.preprocessing.StandardScaler()
    scaler.fit(observation_examples)
    featurizer = sklearn.pipeline.FeatureUnion([
        ("rbf1", RBFSampler(gamma=5.0, n_components=curr_d)),
        ("rbf2", RBFSampler(gamma=2.0, n_components=curr_d)),
        ("rbf3", RBFSampler(gamma=1.0, n_components=curr_d)),
        ("rbf4", RBFSampler(gamma=0.5, n_components=curr_d))
    ])
    featurizer.fit(scaler.transform(observation_examples))

    def feature(s, a):
        tmp = np.zeros(nA * s_dim)
        tmp[a * s_dim: (a + 1) * s_dim] = s
        scaled = scaler.transform([s])
        return featurizer.transform(scaled)[0]

    return feature


def create_random_NN_feature(env, d):
    s_dim = env.observation_space.shape[0]
    nA = env.action_space.n
    net = FeatureNet(s_dim, nA, d)

    def feature(s, a):
        tmp = np.zeros(nA * s_dim)
        a = int(a)
        tmp[a * s_dim: (a + 1) * s_dim] = s
        with torch.no_grad():
            result = net(Variable(torch.tensor([tmp]).float()))[0].numpy()
        return result

    return feature
