from FQE_PG import *
from FQE_PG_utils import *
from Featurizer import *
from PG import PG, sample_gradient, collect_data, PolicyNet, sample_gradient_w
from PG_utils import get_features, EnvWrapper, set_seed
import gym

# initialize
env_name = "FrozenLake-v1"
env = gym.make(env_name)
n_state = env.observation_space.n
n_action = env.action_space.n
env = EnvWrapper(env)

# set seed_init
seed_init = 1000
set_seed(env, seed_init)

# create feature
feature, d = create_tabular_feature(n_state, n_action)

# optimal policy via PG
lr = 0.05
num_episode = 200
N = 100
H = 100
gamma = 1
policy_net = PolicyNet(n_state, n_action)
optimizer = torch.optim.Adam(policy_net.parameters(), lr=lr)
score_his = PG(policy_net, optimizer, env, num_episode, gamma, N, H)

# generate off-policy data: on-policy
K = 200
lam = 0.001
sample_s = 1000
eps_behavior = 0.3
m = net_param_num(policy_net)

FQE_policy_grad_lst, IS_policy_grad_lst = [], []
Sigma_hat_inv_lst = []
seed_lst = [x for x in range(10, 30)]
for seed_i in seed_lst:
    set_seed(env, seed_i)
    states, actions, next_states, rewards = collect_data(env, policy_net, K, H, eps_behavior)
    features = get_features(states, actions, feature)

    # policy value estimation
    FQE_policy_grad, FQE_Q, Sigma_hat_inv = FQE_PG(env, features, next_states, rewards, policy_net, feature, H, d, m, sample_s=sample_s, lam=lam) # FQE_PG_alt
    IS_policy_grad = sample_gradient_w(policy_net, states, actions, rewards,eps_behavior, gamma)
    FQE_policy_grad_lst.append(FQE_policy_grad)
    IS_policy_grad_lst.append(IS_policy_grad)
    Sigma_hat_inv_lst.append(Sigma_hat_inv)

store_result(FQE_policy_grad_lst, "result/alt_FQE_grad_K" + str(K) + "_H" + str(H) + "_lam" + str(lam) + "_eps" + str(eps_behavior) + ".pickle")
store_result(IS_policy_grad_lst, "result/alt_IS_grad_K" + str(K) + "_H" + str(H) + "_lam" + str(lam) + "_eps" + str(eps_behavior) + ".pickle")
store_result(Sigma_hat_inv_lst, "result/alt_Sigma_hat_inv_K" + str(K) + "_H" + str(H) + "_lam" + str(lam) + "_eps" + str(eps_behavior) + ".pickle")
